@@ -1,8 +1,16 @@
-/* es.h -- definitions for higher order shell ($Revision: 1.11 $) */
+/* es.h -- definitions for higher order shell ($Revision: 1.2 $) */
 
 #include "config.h"
 #include "stdenv.h"
 
+/*
+ * meta-information for exported environment strings
+ */
+
+#define ENV_SEPARATOR	'\017'		/* control-O */
+#define	ENV_ESCAPE	'\016'		/* control-N */
+
+
 /*
  * the fundamental es data structures.
  */
@@ -13,11 +21,6 @@ typedef struct List List;
 typedef struct Binding Binding;
 typedef struct Closure Closure;
 
-struct Term {
-	char *str;
-	Closure *closure;
-};
-
 struct List {
 	Term *term;
 	List *next;
@@ -41,7 +44,7 @@ struct Closure {
 
 typedef enum {
 	nAssign, nCall, nClosure, nConcat, nFor, nLambda, nLet, nList, nLocal,
-	nMatch, nPrim, nQword, nThunk, nVar, nVarsub, nWord,
+	nMatch, nExtract, nPrim, nQword, nThunk, nVar, nVarsub, nWord,
 	nRedir, nPipe		/* only appear during construction */
 } NodeKind;
 
@@ -111,9 +114,12 @@ extern void undefer(int ticket);
 /* term.c */
 
 extern Term *mkterm(char *str, Closure *closure);
+extern Term *mkstr(char *str);
 extern char *getstr(Term *term);
 extern Closure *getclosure(Term *term);
 extern Term *termcat(Term *t1, Term *t2);
+extern Boolean termeq(Term *term, const char *s);
+extern Boolean isclosure(Term *term);
 
 
 /* list.c */
@@ -138,15 +144,22 @@ extern Tree *mk(NodeKind VARARGS);
 extern Closure *mkclosure(Tree *tree, Binding *binding);
 extern Closure *extractbindings(Tree *tree);
 extern Binding *mkbinding(char *name, List *defn, Binding *next);
+extern Binding *reversebindings(Binding *binding);
 
 
 /* eval.c */
 
+extern Binding *bindargs(Tree *params, List *args, Binding *binding);
+extern List *forkexec(char *file, List *list, Boolean inchild);
 extern List *walk(Tree *tree, Binding *binding, int flags);
 extern List *eval(List *list, Binding *binding, int flags);
 extern List *eval1(Term *term, int flags);
 extern List *pathsearch(Term *term);
 
+extern unsigned long evaldepth, maxevaldepth;
+#define	MINmaxevaldepth		100
+#define	MAXmaxevaldepth		0xffffffffU
+
 #define	eval_inchild		1
 #define	eval_exitonfalse	2
 #define	eval_flags		(eval_inchild|eval_exitonfalse)
@@ -162,11 +175,13 @@ extern List *glom2(Tree *tree, Binding *binding, StrList **quotep);
 
 extern char QUOTED[], UNQUOTED[];
 extern List *glob(List *list, StrList *quote);
+extern Boolean haswild(const char *pattern, const char *quoting);
 
 
 /* match.c */
 extern Boolean match(const char *subject, const char *pattern, const char *quote);
 extern Boolean listmatch(List *subject, List *pattern, StrList *quote);
+extern List *extractmatches(List *subjects, List *patterns, StrList *quotes);
 
 
 /* var.c */
@@ -174,9 +189,9 @@ extern Boolean listmatch(List *subject, List *pattern, StrList *quote);
 extern void initvars(void);
 extern void initenv(char **envp, Boolean protected);
 extern void hidevariables(void);
-extern char *varname(List *);
-extern List *varlookup(const char *, Binding *);
-extern List *varlookup2(char *name1, char *name2);
+extern void validatevar(const char *var);
+extern List *varlookup(const char *name, Binding *binding);
+extern List *varlookup2(char *name1, char *name2, Binding *binding);
 extern void vardef(char *, Binding *, List *);
 extern Vector *mkenv(void);
 extern void setnoexport(List *list);
@@ -198,12 +213,17 @@ extern char *mkstatus(int status);
 extern void printstatus(int pid, int status);
 
 
+/* access.c */
+
+extern char *checkexecutable(char *file);
+
+
 /* proc.c */
 
 extern Boolean hasforked;
 extern int efork(Boolean parent, Boolean background);
-extern int ewaitfor2(int pid, void *rusage);
-#define	ewaitfor(pid)	ewaitfor2(pid, NULL)
+extern int ewait(int pid, Boolean interruptible, void *rusage);
+#define	ewaitfor(pid)	ewait(pid, FALSE, NULL)
 
 
 /* dict.c */
@@ -245,15 +265,15 @@ extern void sortvector(Vector *v);
 
 /* util.c */
 
-extern char *strerror(int err);
+extern char *esstrerror(int err);
 extern void uerror(char *msg);
 extern void *ealloc(size_t n);
 extern void *erealloc(void *p, size_t n);
 extern void efree(void *p);
-extern char *strdup(const char *s);
 extern void ewrite(int fd, const char *s, size_t n);
 extern long eread(int fd, char *buf, size_t n);
 extern Boolean isabsolute(char *path);
+extern Boolean streq2(const char *s, const char *t1, const char *t2);
 
 
 /* input.c */
@@ -264,6 +284,7 @@ extern Tree *parsestring(const char *str);
 extern void sethistory(char *file);
 extern Boolean isinteractive(void);
 extern void initinput(void);
+extern void resetparser(void);
 
 extern List *runfd(int fd, const char *name, int flags);
 extern List *runstring(const char *str, const char *name, int flags);
@@ -275,6 +296,10 @@ extern List *runstring(const char *str, const char *name, int flags);
 #define	run_printcmds		32	/* -x */
 #define	run_lisptrees		64	/* -L and defined(LISPTREES) */
 
+#if READLINE
+extern Boolean resetterminal;
+#endif
+
 
 /* opt.c */
 
@@ -286,7 +311,7 @@ extern List *esoptend(void);
 
 /* prim.c */
 
-extern List *prim(char *s, List *list, int evalflags);
+extern List *prim(char *s, List *list, Binding *binding, int evalflags);
 extern void initprims(void);
 
 
@@ -295,27 +320,37 @@ extern void initprims(void);
 extern void startsplit(const char *sep, Boolean coalesce);
 extern void splitstring(char *in, size_t len, Boolean endword);
 extern List *endsplit(void);
-extern List *fsplit(const char *sep, List *list);
+extern List *fsplit(const char *sep, List *list, Boolean coalesce);
 
 
 /* signal.c */
 
+extern int signumber(const char *name);
+extern char *signame(int sig);
+extern char *sigmessage(int sig);
+
 #define	SIGCHK() sigchk()
+typedef enum {
+	sig_nochange, sig_catch, sig_default, sig_ignore, sig_noop, sig_special
+} Sigeffect;
+extern Sigeffect esignal(int sig, Sigeffect effect);
+extern void setsigeffects(const Sigeffect effects[]);
+extern void getsigeffects(Sigeffect effects[]);
+extern List *mksiglist(void);
+extern void initsignals(Boolean interactive, Boolean allowdumps);
 extern Atomic slow, interrupted;
 extern jmp_buf slowlabel;
 extern Boolean sigint_newline;
-extern void catcher(int sig);
 extern void sigchk(void);
-extern sigresult (*esignal(int sig, sigresult (*h)(int)))(int);
-extern void initsignals(Boolean allowdumps);
 extern Boolean issilentsignal(List *e);
-extern void setsigdefaults(Boolean background);
-extern void trapsignals(Boolean sigs[]);
+extern void setsigdefaults(void);
+extern void blocksignals(void);
+extern void unblocksignals(void);
 
 
 /* open.c */
 
-typedef enum { oOpen, oCreate, oAppend } OpenKind;
+typedef enum { oOpen, oCreate, oAppend, oReadWrite, oReadCreate, oReadAppend } OpenKind;
 extern int eopen(char *name, OpenKind k);
 
 
@@ -335,8 +370,9 @@ extern char *gcndup(const char *s, size_t n);	/* copy a counted string into gc s
 
 extern void initgc(void);			/* must be called at the dawn of time */
 extern void gc(void);				/* provoke a collection, if enabled */
+extern void gcreserve(size_t nbytes);		/* provoke a collection, if enabled and not enough space */
 extern void gcenable(void);			/* enable collections */
-extern void gcdisable(size_t);			/* disable collections, collect first if space needed */
+extern void gcdisable(void);			/* disable collections */
 extern Boolean gcisblocked();			/* is collection disabled? */
 
 
@@ -352,6 +388,12 @@ struct Root {
 
 extern Root *rootlist;
 
+#if REF_ASSERTIONS
+#define	refassert(e)	assert(e)
+#else
+#define	refassert(e)	NOP
+#endif
+
 #define	Ref(t, v, init) \
 	if (0) ; else { \
 		t v = init; \
@@ -360,8 +402,8 @@ extern Root *rootlist;
 		(CONCAT(v,__root__)).next = rootlist; \
 		rootlist = &(CONCAT(v,__root__))
 #define	RefPop(v) \
-		assert(rootlist == &(CONCAT(v,__root__))); \
-		assert(rootlist->p == (void **) &v); \
+		refassert(rootlist == &(CONCAT(v,__root__))); \
+		refassert(rootlist->p == (void **) &v); \
 		rootlist = rootlist->next;
 #define RefEnd(v) \
 		RefPop(v); \
@@ -377,8 +419,8 @@ extern Root *rootlist;
 		__root__.next = rootlist; \
 		rootlist = &__root__
 #define	RefRemove(e) \
-		assert(rootlist == &__root__); \
-		assert(rootlist->p == (void **) &e); \
+		refassert(rootlist == &__root__); \
+		refassert(rootlist->p == (void **) &e); \
 		rootlist = rootlist->next; \
 	}
 
@@ -412,19 +454,14 @@ struct Push {
 
 
 /*
- * exceptions
- *	typical use is
- *		Handler h;
- *		List *e;
- *		while ((e = pushhandler(&h)) != NULL) {
- *			if (we want to pass e back)
- *				throw(e);
- *			if (we can return)
- *				return ...;
- *			// otherwise just retry
- *		}
- *		... // protected code
- *		pophandler(&h);
+ * exception handling
+ *
+ *	ExceptionHandler
+ *		... body ...
+ *	CatchException (e)
+ *		... catching code using e ...
+ *	EndExceptionHandler
+ *
  */
 
 typedef struct Handler Handler;
@@ -432,21 +469,44 @@ struct Handler {
 	Handler *up;
 	Root *rootlist;
 	Push *pushlist;
+	unsigned long evaldepth;
 	jmp_buf label;
 };
 
 extern Handler *tophandler, *roothandler;
 extern List *exception;
-extern List *pushhandler(Handler *handler);	/* must be a macro */
 extern void pophandler(Handler *handler);
 extern noreturn throw(List *exc);
 extern noreturn fail(const char *from, const char *name VARARGS);
 extern void newchildcatcher(void);
 
-#define	pushhandler(hp)	( \
-		((hp)->rootlist = rootlist), \
-		((hp)->pushlist = pushlist), \
-		((hp)->up = tophandler), \
-		(tophandler = (hp)), \
-		(setjmp((hp)->label) ? exception : NULL) \
-	)
+#if DEBUG_EXCEPTIONS
+extern List *raised(List *e);
+#else
+#define	raised(e)	(e)
+#endif
+
+#define ExceptionHandler \
+	{ \
+		Handler _localhandler; \
+		_localhandler.rootlist = rootlist; \
+		_localhandler.pushlist = pushlist; \
+		_localhandler.evaldepth = evaldepth; \
+		_localhandler.up = tophandler; \
+		tophandler = &_localhandler; \
+		if (!setjmp(_localhandler.label)) {
+	
+#define CatchException(e) \
+			pophandler(&_localhandler); \
+		} else { \
+			List *e = raised(exception); \
+
+#define CatchExceptionIf(condition, e) \
+			if (condition) \
+				pophandler(&_localhandler); \
+		} else { \
+			List *e = raised(exception); \
+
+#define EndExceptionHandler \
+		} \
+	}
