@@ -1,4 +1,4 @@
-/* gc.c -- copying garbage collector for es ($Revision: 1.5 $) */
+/* gc.c -- copying garbage collector for es ($Revision: 1.2 $) */
 
 #define	GARBAGE_COLLECTOR	1	/* for es.h */
 
@@ -18,10 +18,20 @@ struct Space {
 #define	SPACEUSED(sp)	(((sp)->current - (sp)->bot))
 #define	INSPACE(p, sp)	((sp)->bot <= (char *) (p) && (char *) (p) < (sp)->top)
 
+#define	MIN_minspace	10000
+
 #if GCPROTECT
 #define	NSPACES		10
 #endif
 
+#if HAVE_SYSCONF
+# ifndef _SC_PAGESIZE
+#  undef HAVE_SYSCONF
+#  define HAVE_SYSCONF 0
+# endif
+#endif
+
+
 /* globals */
 Root *rootlist;
 int gcblocked = 0;
@@ -33,7 +43,7 @@ static Space *new, *old;
 static Space *spaces;
 #endif
 static Root *globalrootlist;
-static size_t minspace = 10000;		/* minimum number of bytes in a new space */
+static size_t minspace = MIN_minspace;	/* minimum number of bytes in a new space */
 
 
 /*
@@ -41,34 +51,37 @@ static size_t minspace = 10000;		/* minimum number of bytes in a new space */
  */
 
 #if GCVERBOSE
-#define	VERBOSE(p)	if (!gcverbose) ; else eprint p ;
+#define	VERBOSE(p)	STMT(if (gcverbose) eprint p)
 #else
-#define	VERBOSE(p)	do {} while (0)
+#define	VERBOSE(p)	NOP
 #endif
 
 
-#if GCPROTECT
-
 /*
  * GCPROTECT
  *	to use the GCPROTECT option, you must provide the following functions
+ *		initmmu
  *		take
  *		release
  *		invalidate
  *		revalidate
- *		initmmu
  *	for your operating system
  */
 
+#if GCPROTECT
 #if __MACH__
 
-/* mach specific version, for mmu assertion checking */
+/* mach versions of mmu operations */
 
 #include <mach.h>
 #include <mach_error.h>
 
 #define	PAGEROUND(n)	((n) + vm_page_size - 1) &~ (vm_page_size - 1)
 
+/* initmmu -- initialization for memory management calls */
+static void initmmu(void) {
+}
+
 /* take -- allocate memory for a space */
 static void *take(size_t n) {
 	vm_address_t addr;
@@ -110,44 +123,60 @@ static void revalidate(void *p, size_t n) {
 	memset(p, 0x4F, n);
 }
 
-/* initmmu -- initialization for memory management calls */
-static void initmmu(void) {
-}
+#else /* !__MACH__ */
+
+/* sunos-derived mmap(2) version of mmu operations */
 
-#elif sun			/* TODO: sun mmu tweaking */
 #include <sys/mman.h>
 
-extern void mprotect(void *p, size_t nbytes, int flags);
-
 static int pagesize;
 #define	PAGEROUND(n)	((n) + pagesize - 1) &~ (pagesize - 1)
 
 /* take -- allocate memory for a space */
 static void *take(size_t n) {
-	/* TODO: sun take */
+	caddr_t addr;
+#ifdef MAP_ANONYMOUS
+	addr = mmap(0, n, PROT_READ|PROT_WRITE,	MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
+#else
+	static int devzero = -1;
+	if (devzero == -1)
+		devzero = eopen("/dev/zero", oOpen);
+	addr = mmap(0, n, PROT_READ|PROT_WRITE, MAP_PRIVATE, devzero, 0);
+#endif
+	if (addr == (caddr_t) -1)
+		panic("mmap: %s", esstrerror(errno));
+	memset(addr, 0xA5, n);
+	return addr;
 }
 
-/* revalidate -- enable access to a range of memory */
+/* release -- deallocate a range of memory */
 static void release(void *p, size_t n) {
-	/* TODO: sun release */
+	if (munmap(p, n) == -1)
+		panic("munmap: %s", esstrerror(errno));
 }
 
 /* invalidate -- disable access to a range of memory */
 static void invalidate(void *p, size_t n) {
-	mprotect(p, n, PROT_NONE);
+	if (mprotect(p, n, PROT_NONE) == -1)
+		panic("mprotect(PROT_NONE): %s", esstrerror(errno));
 }
 
 /* revalidate -- enable access to a range of memory */
 static void revalidate(void *p, size_t n) {
-	mprotect(p, n, PROT_READ|PROT_WRITE);
+	if (mprotect(p, n, PROT_READ|PROT_WRITE) == -1)
+		panic("mprotect(PROT_READ|PROT_WRITE): %s", esstrerror(errno));
 }
 
 /* initmmu -- initialization for memory management calls */
 static void initmmu(void) {
+#if HAVE_SYSCONF
+	pagesize = sysconf(_SC_PAGESIZE);
+#else
 	pagesize = getpagesize();
+#endif
 }
 
-#endif	/* sun */
+#endif	/* !__MACH__ */
 #endif	/* GCPROTECT */
 
 
@@ -273,7 +302,7 @@ extern void globalroot(void *addr) {
 
 /* not portable to word addressed machines */
 #define	TAG(p)		(((Tag **) p)[-1])
-#define	FORWARDED(tagp)	(((int) tagp) & 1)
+#define	FORWARDED(tagp)	(((size_t) tagp) & 1)
 #define	FOLLOWTO(p)	((Tag *) (((char *) p) + 1))
 #define	FOLLOW(tagp)	((void *) (((char *) tagp) - 1))
 
@@ -307,8 +336,10 @@ extern void *forward(void *p) {
 /* scanroots -- scan a rootlist */
 static void scanroots(Root *rootlist) {
 	Root *root;
-	for (root = rootlist; root != NULL; root = root->next)
+	for (root = rootlist; root != NULL; root = root->next) {
+		VERBOSE(("GC root at %8lx: %8lx\n", root->p, *root->p));
 		*root->p = forward(*root->p);
+	}
 }
 
 /* scanspace -- scan new space until it is up to date */
@@ -347,15 +378,23 @@ extern void gcenable(void) {
 		gc();
 }
 
-/* gcdisable -- disable collections, collect first if space needed */
-extern void gcdisable(size_t minfree) {
+/* gcdisable -- disable collections */
+extern void gcdisable(void) {
 	assert(gcblocked >= 0);
-	if (!gcblocked && SPACEFREE(new) < minfree) {
+	++gcblocked;
+}
+
+/* gcreserve -- provoke a collection if there's not a certain amount of space around */
+extern void gcreserve(size_t minfree) {
+	if (SPACEFREE(new) < (int)minfree) {
 		if (minspace < minfree)
 			minspace = minfree;
 		gc();
 	}
-	++gcblocked;
+#if GCALWAYS
+	else
+		gc();
+#endif
 }
 
 /* gcisblocked -- is collection disabled? */
@@ -424,7 +463,7 @@ extern void gc(void) {
 
 		if (minspace < livedata * 2)
 			minspace = livedata * 4;
-		else if (minspace > livedata * 12)
+		else if (minspace > livedata * 12 && minspace > (MIN_minspace * 2))
 			minspace /= 2;
 
 		--gcblocked;
@@ -484,7 +523,7 @@ DefineTag(String, notstatic);
 extern char *gcndup(const char *s, size_t n) {
 	char *ns;
 
-	gcdisable(0);
+	gcdisable();
 	ns = gcalloc((n + 1) * sizeof (char), &StringTag);
 	memcpy(ns, s, n);
 	ns[n] = '\0';
@@ -595,6 +634,7 @@ static char *tree2name(NodeKind k) {
 	case nList:	return "List";
 	case nLocal:	return "Local";
 	case nMatch:	return "Match";
+	case nExtract:	return "Extract";
 	case nVarsub:	return "Varsub";
 	}
 }
@@ -611,6 +651,7 @@ static char *tree2name(NodeKind k) {
 	};
 
 #include "var.h"
+#include "term.h"
 
 
 static size_t dump(Tag *t, void *p) {
