@@ -1,4 +1,5 @@
-/* input.c -- read input from files or strings ($Revision: 1.6 $) */
+/* input.c -- read input from files or strings ($Revision: 1.2 $) */
+/* stdgetenv is based on the FreeBSD getenv */
 
 #include "es.h"
 #include "input.h"
@@ -8,7 +9,7 @@
  * constants
  */
 
-#define	BUFSIZE		((size_t) 1024)		/* buffer size to fill reads into */
+#define	BUFSIZE		((size_t) 4096)		/* buffer size to fill reads into */
 
 
 /*
@@ -25,14 +26,25 @@
 Input *input;
 char *prompt, *prompt2;
 
-Boolean disablehistory;
+Boolean disablehistory = FALSE;
+Boolean resetterminal = FALSE;
 static char *history;
 static int historyfd = -1;
 
 #if READLINE
-extern int rl_meta_chars;
 extern char *readline(char *);
 extern void add_history(char *);
+extern int read_history(char *);
+extern void stifle_history(int);
+extern void rl_reset_terminal(char *);
+extern char *rl_basic_word_break_characters;
+extern char *rl_completer_quote_characters;
+
+#if ABUSED_GETENV
+static char *stdgetenv(const char *);
+static char *esgetenv(const char *);
+static char *(*realgetenv)(const char *) = stdgetenv;
+#endif
 #endif
 
 
@@ -51,6 +63,11 @@ static char *error = NULL;
 
 /* yyerror -- yacc error entry point */
 extern void yyerror(char *s) {
+#if sgi
+	/* this is so that trip.es works */
+	if (streq(s, "Syntax error"))
+		s = "syntax error";
+#endif
 	if (error == NULL)	/* first error is generally the most informative */
 		error = locate(input, s);
 }
@@ -73,7 +90,7 @@ static void loghistory(const char *cmd, size_t len) {
 	if (historyfd == -1) {
 		historyfd = eopen(history, oAppend);
 		if (historyfd == -1) {
-			eprint("history(%s): %s\n", history, strerror(errno));
+			eprint("history(%s): %s\n", history, esstrerror(errno));
 			vardef("history", NULL, NULL);
 			return;
 		}
@@ -101,97 +118,12 @@ extern void sethistory(char *file) {
 		close(historyfd);
 		historyfd = -1;
 	}
-	history = file;
-}
-
-
-/*
- * getting characters
- */
-
-/* get -- get a character, filter out nulls */
-extern int get(Input *in) {
-	int c;
-	while ((c = (in->buf < in->bufend ? *in->buf++ : (*in->fill)(in))) == '\0')
-		warn("null character ignored");
-	return c;
-}
-
-/* eoffill -- report eof when called to fill input buffer */
-static int eoffill(Input *in) {
-	assert(in->fd == -1);
-	return EOF;
-}
-
-/* callreadline -- signal-safe readline wrapper */
-#if READLINE && !SYSV_SIGNALS
-static char *callreadline(char *prompt) {
-	char *r;
-	interrupted = FALSE;
-	if (!setjmp(slowlabel)) {
-		slow = TRUE;
-		r = interrupted ? NULL : readline(prompt);
-	} else
-		r = NULL;
-	slow = FALSE;
-	if (r == NULL)
-		errno = EINTR;
-	SIGCHK();
-	return r;
-}
-#else
-#define	callreadline readline
-#endif	/* READLINE && !SVSIGS */
-
-/* fdfill -- fill input buffer by reading from a file descriptor */
-static int fdfill(Input *in) {
-	long nread;
-	assert(in->buf == in->bufend);
-	assert(in->fd >= 0);
-
 #if READLINE
-	if (in->runflags & run_interactive) {
-		char *rlinebuf = callreadline(prompt);
-		if (rlinebuf == NULL)
-			nread = 0;
-		else {
-			if (*rlinebuf != '\0')
-				add_history(rlinebuf);
-			nread = strlen(rlinebuf) + 1;
-			if (in->buflen < nread) {
-				while (in->buflen < nread)
-					in->buflen *= 2;
-				efree(in->bufbegin);
-				in->bufbegin = erealloc(in->bufbegin, in->buflen);
-			}
-			memcpy(in->bufbegin, rlinebuf, nread - 1);
-			in->bufbegin[nread - 1] = '\n';
-		}
-	} else
+	/* Attempt to populate readline history with new history file. */
+	stifle_history(50000); /* Keep memory usage within sane-ish bounds. */
+	read_history(file);
 #endif
-	do {
-		nread = eread(in->fd, (char *) in->bufbegin, in->buflen);
-		SIGCHK();
-	} while (nread == -1 && errno == EINTR);
-
-	if (nread <= 0) {
-		close(in->fd);
-		in->fd = -1;
-		in->fill = eoffill;
-		in->runflags &= ~run_interactive;
-		if (nread == -1)
-			fail("$&parse", "%s: %s", in->name == NULL ? "es" : in->name, strerror(errno));
-		return EOF;
-	}
-
-	if (in->runflags & run_echoinput)
-		ewrite(2, (char *) in->bufbegin, nread);
-	if (in->runflags & run_interactive)
-		loghistory((char *) in->bufbegin, nread);
-
-	in->buf = in->bufbegin;
-	in->bufend = &in->buf[nread];
-	return *in->buf++;
+	history = file;
 }
 
 
@@ -220,7 +152,7 @@ extern void unget(Input *in, int c) {
 	if (in->ungot > 0) {
 		assert(in->ungot < MAXUNGET);
 		in->unget[in->ungot++] = c;
-	} else if (in->bufbegin < in->buf && in->buf[-1] == c)
+	} else if (in->bufbegin < in->buf && in->buf[-1] == c && (input->runflags & run_echoinput) == 0)
 		--in->buf;
 	else {
 		assert(in->rfill == NULL);
@@ -236,6 +168,183 @@ extern void unget(Input *in, int c) {
 }
 
 
+/*
+ * getting characters
+ */
+
+/* get -- get a character, filter out nulls */
+static int get(Input *in) {
+	int c;
+	while ((c = (in->buf < in->bufend ? *in->buf++ : (*in->fill)(in))) == '\0')
+		warn("null character ignored");
+	return c;
+}
+
+/* getverbose -- get a character, print it to standard error */
+static int getverbose(Input *in) {
+	if (in->fill == ungetfill)
+		return get(in);
+	else {
+		int c = get(in);
+		if (c != EOF) {
+			char buf = c;
+			ewrite(2, &buf, 1);
+		}
+		return c;
+	}
+}
+
+/* eoffill -- report eof when called to fill input buffer */
+static int eoffill(Input *in) {
+	assert(in->fd == -1);
+	return EOF;
+}
+
+#if READLINE
+/* callreadline -- readline wrapper */
+static char *callreadline(char *prompt) {
+	char *r;
+	if (prompt == NULL)
+		prompt = ""; /* bug fix for readline 2.0 */
+	if (resetterminal) {
+		rl_reset_terminal(NULL);
+		resetterminal = FALSE;
+	}
+	interrupted = FALSE;
+	if (!setjmp(slowlabel)) {
+		slow = TRUE;
+		r = interrupted ? NULL : readline(prompt);
+	} else
+		r = NULL;
+	slow = FALSE;
+	if (r == NULL)
+		errno = EINTR;
+	SIGCHK();
+	return r;
+}
+
+#if ABUSED_GETENV
+
+/* getenv -- fake version of getenv for readline (or other libraries) */
+static char *esgetenv(const char *name) {
+	List *value = varlookup(name, NULL);
+	if (value == NULL)
+		return NULL;
+	else {
+		char *export;
+		static Dict *envdict;
+		static Boolean initialized = FALSE;
+		Ref(char *, string, NULL);
+
+		gcdisable();
+		if (!initialized) {
+			initialized = TRUE;
+			envdict = mkdict();
+			globalroot(&envdict);
+		}
+
+		string = dictget(envdict, name);
+		if (string != NULL)
+			efree(string);
+
+		export = str("%W", value);
+		string = ealloc(strlen(export) + 1);
+		strcpy(string, export);
+		envdict = dictput(envdict, (char *) name, string);
+
+		gcenable();
+		RefReturn(string);
+	}
+}
+
+static char *
+stdgetenv(name)
+	register const char *name;
+{
+	extern char **environ;
+	register int len;
+	register const char *np;
+	register char **p, *c;
+
+	if (name == NULL || environ == NULL)
+		return (NULL);
+	for (np = name; *np && *np != '='; ++np)
+		continue;
+	len = np - name;
+	for (p = environ; (c = *p) != NULL; ++p)
+		if (strncmp(c, name, len) == 0 && c[len] == '=') {
+			return (c + len + 1);
+		}
+	return (NULL);
+}
+
+char *
+getenv(char *name)
+{
+	return realgetenv(name);
+}
+
+extern void
+initgetenv(void)
+{
+	realgetenv = esgetenv;
+}
+
+#endif /* ABUSED_GETENV */
+
+#endif	/* READLINE */
+
+/* fdfill -- fill input buffer by reading from a file descriptor */
+static int fdfill(Input *in) {
+	long nread;
+	assert(in->buf == in->bufend);
+	assert(in->fd >= 0);
+
+#if READLINE
+	if (in->runflags & run_interactive && in->fd == 0) {
+		char *rlinebuf = callreadline(prompt);
+		if (rlinebuf == NULL)
+
+			nread = 0;
+		else {
+			if (*rlinebuf != '\0')
+				add_history(rlinebuf);
+			nread = strlen(rlinebuf) + 1;
+			if (in->buflen < (unsigned int)nread) {
+				while (in->buflen < (unsigned int)nread)
+					in->buflen *= 2;
+				in->bufbegin = erealloc(in->bufbegin, in->buflen);
+			}
+			memcpy(in->bufbegin, rlinebuf, nread - 1);
+			in->bufbegin[nread - 1] = '\n';
+			efree(rlinebuf);
+		}
+	} else
+#endif
+	do {
+		nread = eread(in->fd, (char *) in->bufbegin, in->buflen);
+		SIGCHK();
+	} while (nread == -1 && errno == EINTR);
+
+	if (nread <= 0) {
+		close(in->fd);
+		in->fd = -1;
+		in->fill = eoffill;
+		in->runflags &= ~run_interactive;
+		if (nread == -1)
+			fail("$&parse", "%s: %s", in->name == NULL ? "es" : in->name, esstrerror(errno));
+		return EOF;
+	}
+
+	if (in->runflags & run_interactive)
+		loghistory((char *) in->bufbegin, nread);
+
+	in->buf = in->bufbegin;
+	in->bufend = &in->buf[nread];
+	return *in->buf++;
+}
+
+
 /*
  * the input loop
  */
@@ -249,7 +358,7 @@ extern Tree *parse(char *pr1, char *pr2) {
 	emptyherequeue();
 
 	if (ISEOF(input))
-		throw(mklist(mkterm("eof", NULL), NULL));
+		throw(mklist(mkstr("eof"), NULL));
 
 #if READLINE
 	prompt = (pr1 == NULL) ? "" : pr1;
@@ -259,7 +368,8 @@ extern Tree *parse(char *pr1, char *pr2) {
 #endif
 	prompt2 = pr2;
 
-	gcdisable(300 * sizeof (Tree));
+	gcreserve(300 * sizeof (Tree));
+	gcdisable();
 	result = yyparse();
 	gcenable();
 
@@ -268,7 +378,7 @@ extern Tree *parse(char *pr1, char *pr2) {
 		assert(error != NULL);
 		e = error;
 		error = NULL;
-		fail("$&parse", e);
+		fail("$&parse", "%s", e);
 	}
 #if LISPTREES
 	if (input->runflags & run_lisptrees)
@@ -277,10 +387,17 @@ extern Tree *parse(char *pr1, char *pr2) {
 	return parsetree;
 }
 
+/* resetparser -- clear parser errors in the signal handler */
+extern void resetparser(void) {
+	error = NULL;
+}
+
 /* runinput -- run from an input source */
-extern List *runinput(Input *in, int flags) {
-	Handler h;
-	List *e, *repl, *arg, *result;
+extern List *runinput(Input *in, int runflags) {
+	volatile int flags = runflags;
+	List * volatile result = NULL;
+	List *repl, *dispatch;
+	Push push;
 	const char *dispatcher[] = {
 		"fn-%eval-noprint",
 		"fn-%eval-print",
@@ -288,35 +405,40 @@ extern List *runinput(Input *in, int flags) {
 		"fn-%noeval-print",
 	};
 
+	flags &= ~eval_inchild;
 	in->runflags = flags;
+	in->get = (flags & run_echoinput) ? getverbose : get;
 	in->prev = input;
 	input = in;
 
-	if (flags & run_echoinput && in->buf < in->bufend)
-		ewrite(2, (char *) in->buf, in->bufend - in->buf);
+	ExceptionHandler
+
+		dispatch
+	          = varlookup(dispatcher[((flags & run_printcmds) ? 1 : 0)
+					 + ((flags & run_noexec) ? 2 : 0)],
+			      NULL);
+		if (flags & eval_exitonfalse)
+			dispatch = mklist(mkstr("%exit-on-false"), dispatch);
+		varpush(&push, "fn-%dispatch", dispatch);
+
+		repl = varlookup((flags & run_interactive)
+				   ? "fn-%interactive-loop"
+				   : "fn-%batch-loop",
+				 NULL);
+		result = (repl == NULL)
+				? prim("batchloop", NULL, NULL, flags)
+				: eval(repl, NULL, flags);
+
+		varpop(&push);
+
+	CatchException (e)
 
-	if ((e = pushhandler(&h)) != NULL) {
 		(*input->cleanup)(input);
 		input = input->prev;
 		throw(e);
-	}
 
-	arg = varlookup(dispatcher[
-			  ((flags & run_printcmds) ? 1 : 0)
-			+ ((flags & run_noexec)    ? 2 : 0)
-		], NULL);
-	if (flags & eval_exitonfalse)
-		arg = mklist(mkterm("%exit-on-false", NULL), arg);
-	repl = varlookup((flags & run_interactive) ? "fn-%interactive-loop" : "fn-%batch-loop", NULL);
-	if (repl == NULL)
-		result = prim("batchloop", arg, 0);
-	else {
-		if (arg != NULL)
-			repl = append(repl, arg);
-		result = eval(repl, NULL, flags & eval_inchild);
-	}
+	EndExceptionHandler
 
-	pophandler(&h);
 	input = in->prev;
 	(*in->cleanup)(in);
 	return result;
@@ -329,6 +451,7 @@ extern List *runinput(Input *in, int flags) {
 
 /* fdcleanup -- cleanup after running from a file descriptor */
 static void fdcleanup(Input *in) {
+	unregisterfd(&in->fd);
 	if (in->fd != -1)
 		close(in->fd);
 	efree(in->bufbegin);
@@ -344,6 +467,7 @@ extern List *runfd(int fd, const char *name, int flags) {
 	in.fill = fdfill;
 	in.cleanup = fdcleanup;
 	in.fd = fd;
+	registerfd(&in.fd, TRUE);
 	in.buflen = BUFSIZE;
 	in.bufbegin = in.buf = ealloc(in.buflen);
 	in.bufend = in.bufbegin;
@@ -395,24 +519,23 @@ extern List *runstring(const char *str, const char *name, int flags) {
 
 /* parseinput -- turn an input source into a tree */
 extern Tree *parseinput(Input *in) {
-	Handler h;
-	List *e;
-	Tree *result;
+	Tree * volatile result = NULL;
 
 	in->prev = input;
+	in->runflags = 0;
+	in->get = get;
 	input = in;
 
-	if ((e = pushhandler(&h)) != NULL) {
+	ExceptionHandler
+		result = parse(NULL, NULL);
+		if (get(in) != EOF)
+			fail("$&parse", "more than one value in term");
+	CatchException (e)
 		(*input->cleanup)(input);
 		input = input->prev;
 		throw(e);
-	}
+	EndExceptionHandler
 
-	result = parse(NULL, NULL);
-	if (get(in) != EOF)
-		fail("$&parse", "more than one value in term");
-
-	pophandler(&h);
 	input = in->prev;
 	(*in->cleanup)(in);
 	return result;
@@ -473,6 +596,7 @@ extern void initinput(void) {
 	initparse();
 
 #if READLINE
-	rl_meta_chars = 0;
+	rl_basic_word_break_characters=" \t\n\\'`$><=;|&{()}";
+		rl_completer_quote_characters="'";
 #endif
 }
