@@ -1,11 +1,10 @@
-/* proc.c -- process control system calls ($Revision: 1.5 $) */
+/* proc.c -- process control system calls ($Revision: 1.2 $) */
 
 #include "es.h"
 
 /* TODO: the rusage code for the time builtin really needs to be cleaned up */
 
-#include <sys/wait.h>
-#if BSD_LIMITS
+#if HAVE_WAIT3
 #include <sys/time.h>
 #include <sys/resource.h>
 #endif
@@ -18,7 +17,7 @@ struct Proc {
 	int status;
 	Boolean alive, background;
 	Proc *next, *prev;
-#if BSD_LIMITS
+#if HAVE_WAIT3
 	struct rusage rusage;
 #endif
 };
@@ -61,30 +60,30 @@ extern int efork(Boolean parent, Boolean background) {
 			hasforked = TRUE;
 			break;
 		case -1:
-			fail("es:efork", "fork: %s", strerror(errno));
+			fail("es:efork", "fork: %s", esstrerror(errno));
 		}
 	}
 	closefds();
-	setsigdefaults(FALSE);
+	setsigdefaults();
 	newchildcatcher();
 	return 0;
 }
 
-#if BSD_LIMITS
+#if HAVE_WAIT3
 static struct rusage wait_rusage;
 #endif
 
-/* ewait -- a wait wrapper that interfaces with signals */
-static int ewait(int *statusp) {
+/* dowait -- a wait wrapper that interfaces with signals */
+static int dowait(int *statusp) {
 	int n;
 	interrupted = FALSE;
 	if (!setjmp(slowlabel)) {
 		slow = TRUE;
 		n = interrupted ? -2 :
-#if BSD_LIMITS
-			wait3((union wait *) statusp, 0, &wait_rusage);
+#if HAVE_WAIT3
+			wait3((void *) statusp, 0, &wait_rusage);
 #else
-			wait((union wait *) statusp);
+			wait((void *) statusp);
 #endif
 	} else
 		n = -2;
@@ -104,15 +103,15 @@ static void reap(int pid, int status) {
 			assert(proc->alive);
 			proc->alive = FALSE;
 			proc->status = status;
-#if BSD_LIMITS
+#if HAVE_WAIT3
 			proc->rusage = wait_rusage;
 #endif
 			return;
 		}
 }
 
-/* ewaitfor2 -- wait for a specific process to die, or any process if pid == 0 */
-extern int ewaitfor2(int pid, void *rusage) {
+/* ewait -- wait for a specific process to die, or any process if pid == 0 */
+extern int ewait(int pid, Boolean interruptible, void *rusage) {
 	Proc *proc;
 top:
 	for (proc = proclist; proc != NULL; proc = proc->next)
@@ -120,15 +119,22 @@ top:
 			int status;
 			if (proc->alive) {
 				int deadpid;
-				while ((deadpid = ewait(&proc->status)) != pid)
+				int seen_eintr = FALSE;
+				while ((deadpid = dowait(&proc->status)) != pid)
 					if (deadpid != -1)
 						reap(deadpid, proc->status);
-					else if (errno == EINTR)
-						SIGCHK();
+					else if (errno == EINTR) {
+						if (interruptible)
+							SIGCHK();
+						seen_eintr = TRUE;
+					} else if (errno == ECHILD && seen_eintr)
+						/* TODO: not clear on why this is necessary
+						 * (child procs _sometimes_ disappear after SIGINT) */
+						break;
 					else
-						fail("es:ewaitfor2", "wait: %s", strerror(errno));
+						fail("es:ewait", "wait: %s", esstrerror(errno));
 				proc->alive = FALSE;
-#if BSD_LIMITS
+#if HAVE_WAIT3
 				proc->rusage = wait_rusage;
 #endif
 			}
@@ -142,7 +148,7 @@ top:
 			if (proc->background)
 				printstatus(proc->pid, status);
 			efree(proc);
-#if BSD_LIMITS
+#if HAVE_WAIT3
 			if (rusage != NULL)
 				memcpy(rusage, &proc->rusage, sizeof (struct rusage));
 #else
@@ -152,15 +158,16 @@ top:
 		}
 	if (pid == 0) {
 		int status;
-		while ((pid = ewait(&status)) == -1)
-			if (errno == EINTR)
+		while ((pid = dowait(&status)) == -1) {
+			if (errno != EINTR)
+				fail("es:ewait", "wait: %s", esstrerror(errno));
+			if (interruptible)
 				SIGCHK();
-			else
-				fail("es:ewaitfor2", "wait: %s", strerror(errno));
+		}
 		reap(pid, status);
 		goto top;
 	}
-	fail("es:ewaitfor2", "wait: %d is not a child of this shell", pid);
+	fail("es:ewait", "wait: %d is not a child of this shell", pid);
 	NOTREACHED;
 }
 
@@ -171,7 +178,7 @@ PRIM(apids) {
 	Ref(List *, lp, NULL);
 	for (p = proclist; p != NULL; p = p->next)
 		if (p->background && p->alive) {
-			Term *t = mkterm(str("%d", p->pid), NULL);
+			Term *t = mkstr(str("%d", p->pid));
 			lp = mklist(t, lp);
 		}
 	/* TODO: sort the return value, but by number? */
@@ -192,7 +199,7 @@ PRIM(wait) {
 		fail("$&wait", "usage: wait [pid]");
 		NOTREACHED;
 	}
-	return mklist(mkterm(mkstatus(ewaitfor(pid)), NULL), NULL);
+	return mklist(mkstr(mkstatus(ewait(pid, TRUE, NULL))), NULL);
 }
 
 extern Dict *initprims_proc(Dict *primdict) {
