@@ -1,11 +1,17 @@
-/* status.c -- status manipulations ($Revision: 1.2 $) */
+/* status.c -- status manipulations ($Revision: 1.1.1.1 $) */
 
 #include "es.h"
-#include "sigmsgs.h"
+#include "term.h"
 
-static Term trueterm = { "0", NULL }, falseterm = { "1", NULL };
-static List truelist = { &trueterm, NULL }, falselist = { &falseterm, NULL };
-List *true = &truelist, *false = &falselist;
+static const Term
+	trueterm	= { "0", NULL },
+	falseterm	= { "1", NULL };
+static const List
+	truelist	= { (Term *) &trueterm, NULL },
+	falselist	= { (Term *) &falseterm, NULL };
+List
+	*true		= (List *) &truelist,
+	*false		= (List *) &falselist;
 
 /* istrue -- is this status list true? */
 extern Boolean istrue(List *status) {
@@ -48,28 +54,29 @@ extern int exitstatus(List *status) {
 
 /* mkstatus -- turn a unix exit(2) status into a string */
 extern char *mkstatus(int status) {
-	if (status & 0xff) {
-		const char *core = (status & 0x80) ? "+core" : "";
-		int sig = (status & 0x7f);
-		return (sig < NUMOFSIGNALS && *signals[sig].name != '\0')
-			? str("%s%s", signals[sig].name, core)
-			: str("sig%d%s", sig, core);
+	if (SIFSIGNALED(status)) {
+		char *name = signame(STERMSIG(status));
+		if (SCOREDUMP(status))
+			name = str("%s+core", name);
+		return name;
 	}
-	return str("%d", (status >> 8) & 0xff);
+	return str("%d", SEXITSTATUS(status));
 }
 
 /* printstatus -- print the status if we should */
 extern void printstatus(int pid, int status) {
-	if (status & 0xff) {
-		char *msg = ((status & 0x7f) < NUMOFSIGNALS ? signals[status & 0x7f].msg : "");
-		if (pid != 0)
-			eprint("%d: ", pid);
-		if (status & 0x80) {
+	if (SIFSIGNALED(status)) {
+		const char *msg = sigmessage(STERMSIG(status)), *tail = "";
+		if (SCOREDUMP(status)) {
+			tail = "--core dumped";
 			if (*msg == '\0')
-				eprint("core dumped\n");
+				tail += (sizeof "--") - 1;
+		}
+		if (*msg != '\0' || *tail != '\0') {
+			if (pid == 0)
+				eprint("%s%s\n", msg, tail);
 			else
-				eprint("%s--core dumped\n", msg);
-		} else if (*msg != '\0')
-			eprint("%s\n", msg);
+				eprint("%d: %s%s\n", pid, msg, tail);
+		}
 	}
 }
