@@ -1,16 +1,18 @@
-/* stdenv.h -- set up an environment we can use ($Revision: 1.7 $) */
+/* stdenv.h -- set up an environment we can use ($Revision: 1.3 $) */
 
+#include "esconfig.h"
+#ifdef HAVE_SYS_CDEFS_H
+# include <sys/cdefs.h>
+#endif
 
 /*
  * type qualifiers
  */
 
-#if !USE_CONST
-#define	const
-#endif
-
 #if !USE_VOLATILE
-#define	volatile
+# ifndef volatile
+#  define volatile
+# endif
 #endif
 
 
@@ -18,7 +20,7 @@
  * protect the rest of es source from the dance of the includes
  */
 
-#if USE_UNISTD
+#if HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 
@@ -29,7 +31,11 @@
 #include <string.h>
 #include <stddef.h>
 
-#if USE_STDARG
+#if HAVE_MEMORY_H
+#include <memory.h>
+#endif
+
+#if HAVE_STDARG_H
 #include <stdarg.h>
 #else
 #include <varargs.h>
@@ -40,9 +46,10 @@
 #include <signal.h>
 #include <ctype.h>
 
-#if REQUIRE_STAT || REQUIRE_IOCTL
+/* #if REQUIRE_STAT || REQUIRE_IOCTL */
+/* We need sys/types.h for the prototype of gid_t on Linux */
 #include <sys/types.h>
-#endif
+/* #endif */
 
 #if REQUIRE_IOCTL
 #include <sys/ioctl.h>
@@ -53,7 +60,7 @@
 #endif
 
 #if REQUIRE_DIRENT
-#if USE_DIRENT
+#if HAVE_DIRENT_H
 #include <dirent.h>
 typedef struct dirent Dirent;
 #else
@@ -61,10 +68,12 @@ typedef struct dirent Dirent;
 typedef struct direct Dirent;
 #endif
 /* prototypes for XXXdir functions. comment out if necessary */
+#if !HPUX
 extern DIR *opendir(const char *);
 extern Dirent *readdir(DIR *);
 /*extern int closedir(DIR *);*/
 #endif
+#endif
 
 #if REQUIRE_PWD
 #include <pwd.h>
@@ -74,6 +83,27 @@ extern Dirent *readdir(DIR *);
 #include <fcntl.h>
 #endif
 
+/* stdlib */
+#if __GNUC__
+typedef volatile void noreturn;
+#else
+typedef void noreturn;
+#endif
+
+#if STDC_HEADERS
+# include <stdlib.h>
+#else
+extern noreturn exit(int);
+extern noreturn abort(void);
+extern long strtol(const char *num, char **end, int base);
+extern void *qsort(
+	void *base, size_t nmemb, size_t size,
+	int (*compar)(const void *, const void *)
+);
+#endif /* !STDC_HEADERS */
+
+#include <sys/wait.h>
+#include <time.h>
 
 /*
  * things that should be defined by header files but might not have been
@@ -87,6 +117,20 @@ extern Dirent *readdir(DIR *);
 #define	EOF	(-1)
 #endif
 
+/* setjmp */
+
+#if defined sigsetjmp || HAVE_SIGSETJMP
+/* under linux, sigsetjmp and setjmp are both macros 
+ * -- need to undef setjmp to avoid problems
+ */
+# ifdef setjmp
+#  undef setjmp
+# endif
+# define setjmp(buf) sigsetjmp(buf,1)
+# define longjmp(x,y)     siglongjmp(x,y)
+# define jmp_buf     sigjmp_buf
+#endif
+
 
 /*
  * macros
@@ -99,6 +143,14 @@ extern Dirent *readdir(DIR *);
 #define	memzero(dest, count)	memset(dest, 0, count)
 #define	atoi(s)			strtol(s, NULL, 0)
 
+#if SOLARIS
+#define	STMT(stmt)		if (1) { stmt; } else
+#define	NOP			if (1) ; else
+#else
+#define	STMT(stmt)		do { stmt; } while (0)
+#define	NOP			do {} while (0)
+#endif
+
 #if REISER_CPP
 #define CONCAT(a,b)	a/**/b
 #define STRING(s)	"s"
@@ -116,11 +168,6 @@ extern Dirent *readdir(DIR *);
 #undef TRUE
 typedef enum { FALSE, TRUE } Boolean;
 
-#if __GNUC__
-typedef volatile void noreturn;
-#else
-typedef void noreturn;
-#endif
 
 #if USE_SIG_ATOMIC_T
 typedef volatile sig_atomic_t Atomic;
@@ -128,33 +175,32 @@ typedef volatile sig_atomic_t Atomic;
 typedef volatile int Atomic;
 #endif
 
-#if VOID_SIGNALS
-typedef void sigresult;
-#define	SIGRESULT
-#else
-typedef int sigresult;
-#define	SIGRESULT	0
-#endif
+typedef GETGROUPS_T gidset_t;
 
 
 /*
  * variable argument lists
  */
 
-#if USE_STDARG
+#if HAVE_STDARG_H
 
 #define	VARARGS				, ...
 #define	VARARGS1(t1, v1)		(t1 v1, ...)
 #define	VARARGS2(t1, v1, t2, v2)	(t1 v1, t2 v2, ...)
 #define	VA_START(ap, v)			va_start(ap, v)
 
-#else	/* !USE_STDARG */
+#else	/* !HAVE_STDARG_H */
 
 #define	VARARGS
 #define	VARARGS1(t1, v1)		(v1, va_alist) t1 v1; va_dcl
 #define	VARARGS2(t1, v1, t2, v2)	(v1, v2, va_alist) t1 v1; t2 v2; va_dcl
 #define	VA_START(ap, var)		va_start(ap)
 
+/* __va_* are defined by the compiler */
+#define va_start(ap)		__va_start(ap)
+#define va_copy(dest, src)	__va_copy(dest, src)
+#define va_end(ap)		__va_end(ap)
+
 #endif
 
 
@@ -164,31 +210,27 @@ typedef int sigresult;
 
 #if ASSERTIONS
 #define	assert(expr) \
-	do \
+	STMT( \
 		if (!(expr)) { \
 			eprint("%s:%d: assertion failed (%s)\n", \
 				__FILE__, __LINE__, STRING(expr)); \
-			ABORT(); \
+			abort(); \
 		} \
-	while (0)
+	)
 #else
-#define	assert(ignore) do ; while (0)
-#endif
-
-#if NeXT
-#define	ABORT()	asm("trap #15")
-#else
-#define	ABORT()	abort()
+#define	assert(ignore)	NOP
 #endif
 
 enum { UNREACHABLE = 0 };
-#define	NOTREACHED	do assert(UNREACHABLE); while (1)
+
+
+#define	NOTREACHED	STMT(assert(UNREACHABLE))
 
 /*
  * system calls -- can we get these from some standard header uniformly?
  */
 
-#if !USE_UNISTD
+#if !HAVE_UNISTD_H
 extern int chdir(const char *dirname);
 extern int close(int fd);
 extern int dup(int fd);
@@ -216,24 +258,49 @@ extern int stat(const char *, struct stat *);
 #ifdef NGROUPS
 extern int getgroups(int, int *);
 #endif
-#endif	/* !USE_UNISTD */
+#endif	/* !HAVE_UNISTD_H */
 
 
 /*
- * the c library -- these should be in prototypes in standard headers
+ * hacks to present a standard system call interface
  */
 
-/* stdlib */
+#ifdef HAVE_SETSID
+# define setpgrp(a, b)	setsid()
+#else
+#if defined(linux) || defined(__GLIBC__)
+#include "unistd.h"
+#define setpgrp(a, b)	setpgid(a, b)
+#endif
 
-extern noreturn exit(int);
-extern noreturn abort(void);
-extern long strtol(const char *num, char **end, int base);
-extern void *qsort(
-	void *base, size_t nmemb, size_t size,
-	int (*compar)(const void *, const void *)
-);
+#if sgi
+#define	setpgrp(a, b)	BSDsetpgrp(a,b)
+#endif
+
+#if HPUX
+#define	setpgrp(a, b)	setpgrp()
+#endif
+#endif
+
+#if !HAVE_LSTAT
+#define	lstat	stat
+#endif
+
+
+
+/*
+ * macros for picking apart statuses
+ *	we should be able to use the W* forms from <sys/wait.h> but on
+ *	some machines they take a union wait (what a bad idea!) and on
+ *	others an integer.  we just renamed the first letter to s and
+ *	let things be.  on some systems these could just be defined in
+ *	terms of the W* forms.
+ */
+
+#define	SIFSIGNALED(status)	(((status) & 0xff) != 0)
+#define	STERMSIG(status)	((status) & 0x7f)
+#define	SCOREDUMP(status)	((status) & 0x80)
+#define	SIFEXITED(status)	(!SIFSIGNALED(status))
+#define	SEXITSTATUS(status)	(((status) >> 8) & 0xff)
 
-/* setjmp */
 
-extern int setjmp(jmp_buf env);
-extern noreturn longjmp(jmp_buf env, int val);
